#include <string.h>
#include <syslog.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <time.h>
#include <sys/param.h>
#include <utmp.h>
#include <libutil.h>
#include <errno.h>

#include "radius.h"
#include "fsm.h"
#include "ipcp.h"
#include "upap.h"

int rad_logged_in = 0;
char rad_ipaddr[64];
static rad_port = 0;
static char rad_user[256], rad_tty[256], rad_session_id[17];
static char rad_called_id[64], rad_calling_id[64];

int rad_login(char *user, char *port, char *pwd, u_int32_t remote) {
    char *p;
    size_t len;
    int rad_ret;	/* RADIUS return code */
//    struct tm *tm;
    struct utmp utmp;
    struct rad_handle * rad;
    struct stat st;
    extern time_t stime;
    extern u_int32_t nas_ip;

    if (!nas_ip) inet_aton("127.0.0.1", (struct in_addr *)&nas_ip);

    if((port = strrchr(port, '/')) != NULL)
	++port;

    strncpy(rad_user, user, sizeof(rad_user));
    strncpy(rad_tty, port, sizeof(rad_tty));

    /* Try to get Caller ID */

    bzero(rad_called_id, sizeof(rad_called_id));
    bzero(rad_calling_id, sizeof(rad_calling_id));

    if ((p = getenv("CALLER_ID")) && *p)
        strncpy(rad_calling_id, p, sizeof(rad_calling_id));

    if (!*rad_calling_id || !strcmp(rad_calling_id, "none") || !strcmp(rad_calling_id, "?"))
        strcpy(rad_calling_id, "Unknown");

    /* Try to get tty minor number */

    if (fstat(0, &st)) {
	syslog(LOG_ERR, "fstat(stdin): %m");
        rad_port = 0;
    } else {
	rad_port = minor(st.st_rdev);
    }

    /* Set up Called-Station-Id in form "<nas addr>/<nas port>" */

    snprintf(rad_called_id, sizeof(rad_called_id), "%s/%d", ip_ntoa(nas_ip), rad_port);

    /* RADIUS authentication */

    if (!(rad = rad_auth_open())) {
	syslog(LOG_ERR, "rad_auth_open: %d", errno);
	return UPAP_AUTHNAK;
    }

    if (rad_config(rad, NULL)) {
	    syslog(LOG_ERR, "rad_config: %s", rad_strerror(rad));
	    rad_close(rad);
	    return UPAP_AUTHNAK;
    }
    if (rad_create_request(rad, RAD_ACCESS_REQUEST)) {
	    syslog(LOG_ERR, "rad_create_request(auth): %s", rad_strerror(rad));
	    rad_close(rad);
	    return UPAP_AUTHNAK;
    }

    if (rad_put_int(rad, RAD_NAS_IP_ADDRESS, ntohl(nas_ip)) ||
	rad_put_int(rad, RAD_NAS_PORT_TYPE, RAD_ASYNC) ||
	rad_put_int(rad, RAD_NAS_PORT, rad_port) ||
	rad_put_string(rad, RAD_USER_NAME, rad_user) ||
	rad_put_string(rad, RAD_USER_PASSWORD, pwd) ||
	rad_put_int(rad, RAD_SERVICE_TYPE, RAD_FRAMED) ||
	rad_put_int(rad, RAD_FRAMED_PROTOCOL, RAD_PPP) ||
	rad_put_int(rad, RAD_FRAMED_IP_ADDRESS, ntohl(remote)) ||
	rad_put_string(rad, RAD_CALLING_STATION_ID, rad_calling_id) ||
	rad_put_string(rad, RAD_CALLED_STATION_ID, rad_called_id)) {
	syslog(LOG_ERR, "rad_put_XXX: %s", rad_strerror(rad));
	rad_close(rad);
	return UPAP_AUTHNAK;
    }

    if ((rad_ret = rad_send_request(rad)) == -1) {
	    syslog(LOG_ERR, "rad_send_request(auth): %s", rad_strerror(rad));
	    rad_close(rad);
	    return UPAP_AUTHNAK;
    }

    if (rad_ret != RAD_ACCESS_ACCEPT) {
	    syslog(LOG_INFO, "authorization failure for %s: %s", rad_user, rad_strerror(rad));
	    rad_close(rad);
	    return UPAP_AUTHNAK;
    }

    while ((rad_ret = rad_get_attr(rad, (const void **) &p, &len)) && rad_ret != -1) {
	    if (rad_ret == RAD_FRAMED_IP_ADDRESS) {
		remote = htonl(rad_cvt_int(p));
	    }
    }

    rad_close(rad);	/* RADIUS AUTH request finished successfully */

    /* Save Framed-Ip-Address for further processing */

    strncpy(rad_ipaddr, ip_ntoa(remote), sizeof(rad_ipaddr));

    /* Create Session-ID */
    /* FIXME: try to md5(nas_ip, stime, pid) */

    sprintf(rad_session_id, "%08X%08X", stime, getpid());

    /* RADIUS accounting */

    if (!(rad = rad_acct_open())) {
	syslog(LOG_ERR, "rad_acct_open: %d", errno);
	return UPAP_AUTHNAK;
    }

    if (rad_config(rad, NULL)) {
	    syslog(LOG_ERR, "rad_config: %s", rad_strerror(rad));
	    rad_close(rad);
	    return UPAP_AUTHNAK;
    }

    if (rad_create_request(rad, RAD_ACCOUNTING_REQUEST)) {
	    syslog(LOG_ERR, "rad_create_request(acct): %s", rad_strerror(rad));
	    rad_close(rad);
	    return UPAP_AUTHNAK;
    }

    if (rad_put_int(rad, RAD_NAS_IP_ADDRESS, ntohl(nas_ip)) ||
	rad_put_int(rad, RAD_NAS_PORT_TYPE, RAD_ASYNC) ||
	rad_put_int(rad, RAD_NAS_PORT, rad_port) ||
	rad_put_string(rad, RAD_USER_NAME, rad_user) ||
	rad_put_int(rad, RAD_SERVICE_TYPE, RAD_FRAMED) ||
	rad_put_int(rad, RAD_FRAMED_PROTOCOL, RAD_PPP) ||
	rad_put_int(rad, RAD_FRAMED_IP_ADDRESS, htonl(remote)) ||
	rad_put_string(rad, RAD_CALLING_STATION_ID, rad_calling_id) ||
	rad_put_string(rad, RAD_CALLED_STATION_ID, rad_called_id) ||
	rad_put_string(rad, RAD_ACCT_SESSION_ID, rad_session_id) ||
	rad_put_int(rad, RAD_ACCT_STATUS_TYPE, RAD_START)) {
	syslog(LOG_ERR, "rad_put_XXX: %s", rad_strerror(rad));
	rad_close(rad);
	return UPAP_AUTHNAK;
    }

    if ((rad_ret = rad_send_request(rad)) == -1) {
	    syslog(LOG_ERR, "rad_send_request(acct): %s", rad_strerror(rad));
	    rad_close(rad);
	    return UPAP_AUTHNAK;
    }

    rad_close(rad);	/* RADIUS ACCT request finished successfully */

    bzero(&utmp, sizeof(utmp));
    time(&utmp.ut_time);
    strncpy(utmp.ut_name, rad_user, sizeof(utmp.ut_name));
    utmp.ut_host[0]='*';
    strncpy(utmp.ut_host+1, rad_ipaddr, sizeof(utmp.ut_host)-1);
    strncpy(utmp.ut_line, rad_tty, sizeof(utmp.ut_line));
    logwtmp(rad_tty, "", "");
    logout(rad_tty);
    login(&utmp);

    syslog(LOG_INFO, "user %s logged in", rad_user);
    rad_logged_in = 1;
    return UPAP_AUTHACK;
}

void rad_logout() {
    struct rad_handle * rad;
    int rad_ret;	/* RADIUS return code */
    extern time_t stime;
    time_t etime = time((time_t *) NULL);
    extern u_int32_t nas_ip;
    ipcp_options *ho = &ipcp_hisoptions[0];	/* it's safe cause NUM_PPP == 1 */

    /* RADIUS accounting */

    if (!(rad = rad_acct_open())) {
	syslog(LOG_ERR, "rad_acct_open: %d", errno);
	goto END;
    }

    if (rad_config(rad, NULL)) {
	    syslog(LOG_ERR, "rad_config: %s", rad_strerror(rad));
	    rad_close(rad);
	    goto END;
    }

    if (rad_create_request(rad, RAD_ACCOUNTING_REQUEST)) {
	    syslog(LOG_ERR, "rad_create_request(acct): %s", rad_strerror(rad));
	    rad_close(rad);
	    goto END;
    }

    if (rad_put_int(rad, RAD_NAS_IP_ADDRESS, ntohl(nas_ip)) ||
	rad_put_int(rad, RAD_NAS_PORT_TYPE, RAD_ASYNC) ||
	rad_put_int(rad, RAD_NAS_PORT, rad_port) ||
	rad_put_string(rad, RAD_USER_NAME, rad_user) ||
	rad_put_int(rad, RAD_SERVICE_TYPE, RAD_FRAMED) ||
	rad_put_int(rad, RAD_FRAMED_PROTOCOL, RAD_PPP) ||
	rad_put_int(rad, RAD_FRAMED_IP_ADDRESS, ntohl(inet_addr(rad_ipaddr))) ||
	rad_put_int(rad, RAD_ACCT_SESSION_TIME, etime - stime) ||
	rad_put_string(rad, RAD_CALLING_STATION_ID, rad_calling_id) ||
	rad_put_string(rad, RAD_CALLED_STATION_ID, rad_called_id) ||
	rad_put_string(rad, RAD_ACCT_SESSION_ID, rad_session_id) ||
	rad_put_int(rad, RAD_ACCT_STATUS_TYPE, RAD_STOP)) {
	syslog(LOG_ERR, "rad_put_XXX: %s", rad_strerror(rad));
	rad_close(rad);
	goto END;
    }

    if ((rad_ret = rad_send_request(rad)) == -1) {
	    syslog(LOG_ERR, "rad_send_request(acct): %s", rad_strerror(rad));
	    rad_close(rad);
	    goto END;
    }

    rad_close(rad);	/* RADIUS ACCT request finished successfully */

END:

    logwtmp(rad_tty, "", "");	/* wipe out wtmp log entry */
    logout(rad_tty);		/* wipe out utmp log entry */

    rad_logged_in = 0;
}
